{% extends "!layout.html" %}


{% block rootrellink %}
        <li><a href="{{ pathto('intro') }}">home</a>|&nbsp;</li>
        <li><a href="{{ pathto('search') }}">search</a>|&nbsp;</li>
        <!-- 
	<li><a href="{{ pathto('contents') }}">documentation </a> &raquo;</li>
	-->
{% endblock %}


{% block relbar1 %}

<div style="background-color: white; text-align: left; padding: 10px 10px 15px 15px">
<a href="{{ pathto('index') }}"><img src="{{
pathto("_static/logo.png", 1) }}" border="0" alt="sampledoc"/></a>
</div>
{{ super() }}
{% endblock %}

{# put the sidebar before the body #}
{% block sidebar1 %}{{ sidebar() }}{% endblock %}
{% block sidebar2 %}{% endblock %}

{#
{% block footer %}
        <a rel="license" href="http://creativecommons.org/licenses/by-nc-sa/4.0/"><img alt="Creative Commons License" style="border-width:0" src="https://i.creativecommons.org/l/by-nc-sa/4.0/88x31.png" /></a><br /><span xmlns:dct="http://purl.org/dc/terms/" property="dct:title">climlab Documentation</span> by <span xmlns:cc="http://creativecommons.org/ns#" property="cc:attributionName">Moritz Kreuzer</span> is licensed under a <a rel="license" href="http://creativecommons.org/licenses/by-nc-sa/4.0/">Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License</a>.<br />Based on a work at <a xmlns:dct="http://purl.org/dc/terms/" href="https://github.com/climlab/climlab" rel="dct:source">https://github.com/climlab/climlab</a>.
{% endblock %}
#}

{#
{% block footer %}
<a rel="license" href="http://creativecommons.org/licenses/by-nc-sa/4.0/"><img alt="Creative Commons License" style="border-width:0" src="https://i.creativecommons.org/l/by-nc-sa/4.0/80x15.png" /></a>, Moritz Kreuzer. Created using <a xmlns:dct="http://purl.org/dc/terms/" href="http://www.sphinx-doc.org/" rel="dct:source">Sphinx</a>
{% endblock %}
#}
