rm *.pyc
rm climlab/*.pyc
rm climlab/radiation/*.pyc
rm climlab/radiation/cam3/*.pyc
rm climlab/radiation/rrtm/*.pyc
rm climlab/radiation/rrtm/_rrtmg_lw/*.pyc
rm climlab/radiation/rrtm/_rrtmg_sw/*.pyc
rm -r .cache
rm -r build
rm climlab/radiation/rrtm/_rrtmg_lw/*.so
rm -r climlab/radiation/rrtm/_rrtmg_lw/*.so.dSYM
rm climlab/radiation/rrtm/_rrtmg_sw/*.so
rm -r climlab/radiation/rrtm/_rrtmg_sw/*.so.dSYM
rm climlab/radiation/cam3/*.so
rm -r climlab/radiation/cam3/*.so.dSYM
rm -r climlab.egg-info
