'''A collection of general purpose utility functions.'''

# Root URL for the climlab data repository
_datapath_http = "http://www.atmos.albany.edu/facstaff/brose/resources/climlab_data/"
# Alternate path that should also work
_threddspath_http = "http://thredds.atmos.albany.edu:8080/thredds/fileServer/CLIMLAB/"
